## ----style, echo = FALSE, results = 'hide'------------------------------------
BiocStyle::markdown()

## -----------------------------------------------------------------------------
# if (!require("BiocManager", quietly = TRUE))
#     install.packages("BiocManager")
# 
# BiocManager::install("CytoMDS")

## ----libs, results = FALSE----------------------------------------------------
library(CytoPipeline)
library(CytoMDS)
library(ggplot2)

## ----OMIP021Trans-------------------------------------------------------------
data(OMIP021Samples)

#outputDir <- base::tempdir()

transList <- estimateScaleTransforms(
    ff = OMIP021Samples[[1]],
    fluoMethod = "estimateLogicle",
    scatterMethod = "linearQuantile",
    scatterRefMarker = "BV785 - CD3")

OMIP021Trans <- CytoPipeline::applyScaleTransforms(
    OMIP021Samples[,c(1:16,18,20:22)], # removing 'EMPTY' channels
    transList)

## ----OMIPSim1-----------------------------------------------------------------
nSample <- 20
ffList <- list()

# to allow for reproducibility of sample()
set.seed(0)

for (i in seq_len(nSample)) {
    ffList[[i]] <- CytoPipeline::subsample(
                OMIP021Trans[[(i+1)%%2+1]],
                nEvents = 1000,
                seed = i)
}
OMIP021Sim1 <- as(ffList, "flowSet")

pData1 <- data.frame(name = paste0("S", seq_len(nSample)),
                     original_sample = factor(rep(c(1,2), 5)))

## ----OMIPSim2-----------------------------------------------------------------
nSample <- 20
ffList <- list()
S1prop <- rep(0., nSample)

# to allow for reproducibility of sample()
set.seed(0)

for (i in seq_len(nSample)) {
    if (i==1) {
        ffList[[i]] <- OMIP021Trans[[1]]
        S1prop[i] <- 1
    } else if (i==nSample) {
        ffList[[i]] <- OMIP021Trans[[2]]
        S1prop[i] <- 0
    } else {
        ff1 <- CytoPipeline::subsample(
            OMIP021Trans[[1]],
            nEvents = 1000 * (nSample-i)/(nSample-1),
            seed = i)
        ff2 <- CytoPipeline::subsample(
            OMIP021Trans[[2]],
            nEvents = 1000 * (i-1)/(nSample-1),
            seed = i)
        S1prop[i] <- (nSample-i)/(nSample-1)
        
        ffList[[i]] <- CytoPipeline::aggregateAndSample(
            flowCore::flowSet(ff1, ff2),
            nTotalEvents = 1000000 # big number to have a simple aggregation
        )[,1:20]
    }
}
OMIP021Sim2 <- as(ffList, "flowSet")

pData2 <- data.frame(name = paste0("S", seq_len(nSample)),
                     origin = factor(c("Raw", rep("Sim", nSample-2), "Raw")),
                     rawLabel = c("D1", rep("", nSample-2), "D2"),
                     S1prop = S1prop)


## ----distCalc-----------------------------------------------------------------
pwDist1 <- pairwiseEMDDist(
    OMIP021Sim1, channels = NULL)

pwDist2 <- pairwiseEMDDist(
    OMIP021Sim2, channels = NULL)

## ----dist1Show----------------------------------------------------------------
round(pwDist1[1:10, 1:10], 2)

## ----dist1ShowHist, fig.align='center', fig.wide = TRUE-----------------------
distVec1 <- pwDist1[upper.tri(pwDist1)]
distVecDF1 <- data.frame(dist = distVec1)
pHist1 <- ggplot(distVecDF1, mapping = aes(x=dist)) + 
    geom_histogram(fill = "darkgrey", col = "black", bins = 15) + 
    theme_bw() + ggtitle("EMD distances for data set 1")
pHist1

## ----dist12ShowHist, fig.align='center', fig.wide = TRUE----------------------
distVec2 <- pwDist2[upper.tri(pwDist2)]
distVecDF2 <- data.frame(dist = distVec2)
ggplot(distVecDF2, mapping = aes(x=dist)) + 
    geom_histogram(bins = 15, fill = "darkgrey", col = "black") + 
    theme_bw() + ggtitle("EMD distances for data set 2")

## ----MDSCalc1-----------------------------------------------------------------
mdsObj1 <- CytoMDS::computeMetricMDS(pwDist1, seed = 0)

## ----plotMDS1_simplest, fig.align='center', fig.asp = 1, fig.wide = TRUE------
ggplotSampleMDS(mdsObj1)

## ----plotMDS1_colors_shapes, fig.align='center', fig.asp = 0.8, fig.wide = TRUE----
ggplotSampleMDS(mdsObj1, 
                pData = pData1, 
                pDataForColour = "original_sample",
                pDataForShape = "original_sample",
                pDataForLabel = "name")

## ----plotMDSShepard, fig.align='center', fig.wide = TRUE, fig.asp = 1---------
ggplotSampleMDSShepard(mdsObj1)

## ----MDSCalc2.noOption, fig.align='center', fig.asp = 0.85, fig.wide = TRUE----
mdsObj2 <- CytoMDS::computeMetricMDS(pwDist2)
ggplotSampleMDS(mdsObj2,
                pData = pData2, 
                pDataForLabel = "rawLabel",
                pDataForShape = "origin",
                pDataForColour = "S1prop")

## ----MDSCalc2.axes3.4, fig.align='center', fig.asp = 0.85, fig.wide = TRUE----
ggplotSampleMDS(mdsObj2,
                projectionAxes = c(2, 3),
                pData = pData2, 
                pDataForLabel = "rawLabel",
                pDataForShape = "origin",
                pDataForColour = "S1prop")

## ----MDSCalc2.nDim2, fig.align='center', fig.asp = 0.85, fig.wide = TRUE------
mdsObj2_2 <- CytoMDS::computeMetricMDS(pwDist2, nDim = 2)
ggplotSampleMDS(mdsObj2_2,
                pData = pData2, 
                pDataForLabel = "rawLabel",
                pDataForShape = "origin",
                pDataForColour = "S1prop")

## ----MDSCalc2.Rsq99, fig.align='center', fig.asp = 0.85, fig.wide = TRUE------
mdsObj2_3 <- CytoMDS::computeMetricMDS(pwDist2, targetPseudoRSq = 0.99)
ggplotSampleMDS(mdsObj2_3,
                pData = pData2, 
                pDataForLabel = "rawLabel",
                pDataForShape = "origin",
                pDataForColour = "S1prop")

## ----biplot median, fig.align='center', fig.wide = TRUE, fig.asp = 1----------
ggplotSampleMDSShepard(mdsObj2_3)

## ----plotMDS1_medians---------------------------------------------------------
medians <- channelSummaryStats(OMIP021Sim1, statFUNs = median)

## ----plotMDS1_biplot, fig.align='center', fig.asp = 0.8, fig.wide = TRUE------
ggplotSampleMDS(mdsObj1, 
                pData = pData1, 
                pDataForColour = "original_sample",
                pDataForShape = "original_sample",
                displayPointLabels = FALSE,
                displayArrowLabels = TRUE,
                repelArrowLabels = TRUE,
                biplot = TRUE,
                extVariables = medians)

## ----plotMDS1_biplot_0.9, fig.align='center', fig.asp = 0.8, fig.wide = TRUE----
ggplotSampleMDS(mdsObj1, 
                pData = pData1, 
                pDataForColour = "original_sample",
                pDataForShape = "original_sample",
                displayPointLabels = FALSE,
                displayArrowLabels = TRUE,
                repelArrowLabels = TRUE,
                biplot = TRUE,
                extVariables = medians,
                arrowThreshold = 0.9) 

## ----plotMDS1_stats-----------------------------------------------------------
statFUNs = c("median" = stats::median,
             "Q25" = function(x, na.rm) {
                 stats::quantile(x, probs = 0.25)
             },
             "Q75" = function(x, na.rm) {
                 stats::quantile(x, probs = 0.75)
             },
             "standard deviation" = stats::sd)
chStats <- channelSummaryStats(OMIP021Sim1, statFUNs = statFUNs)

## ----plotMDS1_biplot_facetting, fig.align='center', fig.asp = 1, fig.wide = TRUE----
ggplotSampleMDSWrapBiplots(
    mdsObj1, 
    extVariableList = chStats,
    ncol = 2,
    pData = pData1,
    pDataForColour = "original_sample",
    pDataForShape = "original_sample",
    displayPointLabels = FALSE,
    displayArrowLabels = TRUE,
    repelArrowLabels = TRUE,
    arrowThreshold = 0.9,
    displayLegend = FALSE) 

## ----plotMDS1_biplot_stddev, fig.align='center', fig.asp = 2.5, fig.wide = TRUE----
stdDevs <- list()
stdDevs[["std dev of channels 1 to 6"]] <- 
    chStats[["standard deviation"]][,1:6]
stdDevs[["std dev of channels 7 to 12"]] <- 
            chStats[["standard deviation"]][,7:12]
stdDevs[["std dev of channels 13 to 18"]] <- 
            chStats[["standard deviation"]][,13:18]
ggplotSampleMDSWrapBiplots(
    mdsObj1, 
    ncol = 1,
    extVariableList = stdDevs,
    pData = pData1, 
    pDataForColour = "original_sample",
    pDataForShape = "original_sample",
    displayPointLabels = FALSE,
    displayArrowLabels = TRUE,
    repelArrowLabels = TRUE,
    arrowThreshold = 0.9) 

## ----temporary_store----------------------------------------------------------
storageLocation <- suppressMessages(base::tempdir())

nSample <- length(OMIP021Sim1)
fileNames <- file.path(
    storageLocation,
    paste0("OMIP021Sim1_Sample", seq_len(nSample), ".rds"))

for (i in seq_len(nSample)) {
    saveRDS(OMIP021Sim1[[i]], 
            file = fileNames[i])
}

## ----distance_calc_loading_otf------------------------------------------------
pwDist1Again <- pairwiseEMDDist(
    x = nSample,
    loadFlowFrameFUN = function(ffIndex, theFiles){
        readRDS(file = theFiles[ffIndex])
    },
    loadFlowFrameFUNArgs = list(theFiles = fileNames),
    verbose = TRUE
)

## ----distance_calc_loading_otf_parallel---------------------------------------
bp <- BiocParallel::SnowParam(
    stop.on.error = FALSE,
    progressbar = TRUE)
pwDist1Last <- suppressWarnings(pairwiseEMDDist(
    x = nSample,
    loadFlowFrameFUN = function(ffIndex, theFiles){
        readRDS(file = theFiles[ffIndex])
    },
    loadFlowFrameFUNArgs = list(theFiles = fileNames),
    useBiocParallel = TRUE,
    BPPARAM = bp))

## ----dist1ShowHistAgain, fig.align='center', fig.wide = TRUE------------------
distVec1Last <- pwDist1Last[upper.tri(pwDist1Last)]
distVecDF1Last <- data.frame(dist = distVec1Last)
pHist1Last <- ggplot(distVecDF1Last, mapping = aes(x=dist)) + 
    geom_histogram(fill = "darkgrey", col = "black", bins = 15) + 
    theme_bw() + ggtitle("EMD distances for data set 1 - parallel computation")
pHist1Last

## ----sessioninfo, echo=FALSE--------------------------------------------------
sessionInfo()

